from minigrid.core.grid import Grid
from minigrid.core.mission import MissionSpace
from minigrid.core.world_object import Goal, Lava
from minigrid.minigrid_env import MiniGridEnv, WorldObj, COLORS, fill_coords, point_in_rect


class BlueGoal(WorldObj):
    def __init__(self):
        super().__init__("goal", "blue")

    def can_overlap(self):
        return True

    def render(self, img):
        fill_coords(img, point_in_rect(0, 1, 0, 1), COLORS[self.color])


class DeceivingRewardsEnv(MiniGridEnv):
    """
    ### Description

    Environment with 2 goals, one easy to get and one requiring more exploration.
    The agent must find the access to a room to get the second reward.

    ### Mission Space

    "get to a green goal square"

    ### Action Space

    | Num | Name         | Action       |
    |-----|--------------|--------------|
    | 0   | left         | Turn left    |
    | 1   | right        | Turn right   |
    | 2   | forward      | Move forward |
    | 3   | pickup       | Unused       |
    | 4   | drop         | Unused       |
    | 5   | toggle       | Unused       |
    | 6   | done         | Unused       |

    ### Observation Encoding

    - Each tile is encoded as a 3 dimensional tuple:
        `(OBJECT_IDX, COLOR_IDX, STATE)`
    - `OBJECT_TO_IDX` and `COLOR_TO_IDX` mapping can be found in
        [minigrid/minigrid.py](minigrid/minigrid.py)
    - `STATE` refers to the door state with 0=open, 1=closed and 2=locked

    ### Rewards

    A reward of '1' is given for the easy goal, and '0.5' for the hard goal.

    ### Termination

    The episode ends if any one of the following conditions is met:

    1. The agent reaches the goal.
    2. Timeout (see `max_steps`).

    ### Registered Configurations

    - `MiniGrid-DeceivingRewards-v0`
    """

    def __init__(self, **kwargs):

        mission_space = MissionSpace(
            mission_func=lambda: "get to the green goal square"
        )

        size = 20
        super().__init__(
            mission_space=mission_space,
            grid_size=size,
            max_steps=4 * size * size,
            see_through_walls=True,  # Set this to True for maximum speed
            **kwargs
        )

    def _gen_grid(self, width, height):

        # Create an empty grid
        self.grid = Grid(width, height)

        # Generate the surrounding walls
        self.grid.wall_rect(0, 0, width, height)

        room_w = width // 2
        room_h = height // 2

        # For each row of rooms
        for j in range(0, 2):

            # For each column
            for i in range(0, 2):

                if j != 1 and i != 1:
                    continue

                xL = i * room_w
                yT = j * room_h
                xR = xL + room_w
                yB = yT + room_h

                # Bottom wall and door
                if i + 1 < 2:
                    self.grid.vert_wall(xR, yT, room_h)
                    pos = (xR, yB - 2)
                    self.grid.set(*pos, None)

                # Bottom wall and door
                if j + 1 < 2:
                    self.grid.horz_wall(xL, yB, room_w)

        # Place a green goal square in the upper-right corner
        self.put_obj(Goal(), width - 2, 1)

        # Place a blue goal square in the bottom-right corner
        self.put_obj(BlueGoal(), width - 2, height - 2)

        # Place some lava to make goal more distinguishable
        self.grid.vert_wall(height - 2, room_h + 1, 6, Lava)

        # Place the agent
        self.agent_pos = (width - 5, room_h - 3)
        self.agent_dir = 3

        self.mission = "get to the green goal square"

    def _reward(self):
        """
        Compute the reward to be given upon success
        """
        return 1



import torch


def halfcheetah_bullet(state: torch.Tensor, action: torch.Tensor, next_state: torch.Tensor) -> torch.Tensor:
    """   
    HalfCheetahBulletEnv-v0 velocity is 3 idx:
    https://github.com/bulletphysics/bullet3/blob/478da7469a34074aa051e8720734287ca371fd3e/examples/pybullet/gym/pybullet_envs/robot_locomotors.py#L64
    """
    x_velocities = state[:, 3]
    action_penalty = - 0.1 * (torch.sum(action**2, axis=1))
    reward = (x_velocities + action_penalty)[:, None]
    assert reward.shape == (state.shape[0], 1)
    return reward


import random
from animalai.envs.arena_config import Vector3


def analyze_arena(arena):
    tot_reward = 0
    max_good = 0
    goods = []
    goodmultis = []
    for i in arena.arenas[0].items:
        if i.name in ['GoodGoal', 'GoodGoalBounce']:
            if len(i.sizes) == 0:  # arenas max cannot be computed
                return -1
            max_good = max(i.sizes[0].x, max_good)
            goods.append(i.sizes[0].x)
        if i.name in ['GoodGoalMulti', 'GoodGoalMultiBounce']:
            if len(i.sizes) == 0:  # arenas max cannot be computed
                return -1
            tot_reward += i.sizes[0].x
            goodmultis.append(i.sizes[0].x)

    tot_reward += max_good
    goods.sort()
    goodmultis.sort()
    return tot_reward


def random_size_reward():
    # according to docs it's 0.5-5
    s = random.randint(5, 50) / 10
    return (s, s, s)


def set_reward_arena(arena, force_new_size=False):
    tot_reward = 0
    max_good = 0
    goods = []
    goodmultis = []
    for i in arena.arenas[0].items:
        if i.name in ['GoodGoal', 'GoodGoalBounce']:
            if len(i.sizes) == 0 or force_new_size:
                x, y, z = random_size_reward()
                i.sizes = []  # remove previous size if there
                i.sizes.append(Vector3(x, y, z))
            max_good = max(i.sizes[0].x, max_good)
            goods.append(i.sizes[0].x)
        if i.name in ['GoodGoalMulti', 'GoodGoalMultiBounce']:
            if len(i.sizes) == 0 or force_new_size:
                x, y, z = random_size_reward()
                i.sizes = []  # remove previous size if there
                i.sizes.append(Vector3(x, y, z))
            tot_reward += i.sizes[0].x
            goodmultis.append(i.sizes[0].x)

    tot_reward += max_good
    goods.sort()
    goodmultis.sort()
    return tot_reward


import torch
import numpy as np
from torch.distributions.kl import kl_divergence

import pytorchrl as prl
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn


class AttractionKL(PolicyLossAddOn):

    def __init__(self,
                 behavior_factories,
                 behavior_weights,
                 loss_term_weight=1.0,
                 eps=1e-8):
        """
        Class to enforce similarity of any algorithm policy to specified list of behaviors.
        We use the same loss term as in https://arxiv.org/pdf/2105.12196.pdf.

        Parameters
        ----------
        behavior_factories : list
            List of methods creating the agent behaviors.
        behavior_weights : list
            List of floats giving relative weight to each agent behavior. All weights should be
            positive. Otherwise AssertionError will be raised.
        loss_term_weight : float
            Weight of the KL term in the algorithm policy loss.
        eps : float
            Lower bound for prob values, used to clip action probs.
        """

        # Check sizes match
        assert len(behavior_factories) == len(behavior_weights)

        self.eps = eps
        self.behaviors = []
        self.loss_term_weight = loss_term_weight
        self.behavior_factories = behavior_factories

        # Check all behavior weights are positive
        assert (np.array(behavior_weights) >= 0.0).all()

        # Normalize behavior_weights
        self.behavior_weights = behavior_weights
        self.behavior_weights /= np.sum(self.behavior_weights)

    def setup(self, actor, device):
        """
        Setup addon module by casting behavior weights to torch tensors and
        initializing agent behaviors.
        """

        self.actor = actor
        self.device = device

        # Cast behavior weights to torch tensors
        self.behavior_weights = [torch.tensor(w).to(device) for w in self.behavior_weights]

        # Create behavior instances
        for b in self.behavior_factories:
            self.behaviors.append(b(self.device))

    def compute_loss_term(self, data, actor_dist, info):
        """
        Calculate and add KL Attraction loss term.
            1. Calculate KL between actor policy and all behaviors.
            2. Compute biased KL similarities and select minimum value.
            3. Multiply the result by the loss_term_weight.
            4. Change sign of the loss term so KL between behaviors is minimized.

        Parameters
        ----------
        actor_dist : torch.distributions.Distribution
            Actor action distribution for actions in data[prl.OBS]
        data : dict
            data batch containing all required tensors to compute loss term.
        info : dict
            Dictionary to store log information.

        Returns
        -------
        attraction_kl_loss_term : torch.tensor
            KL loss term.
        info : dict
            Updated info dict.

        """

        o, rhs, a, d = data[prl.OBS], data[prl.RHS], data[prl.ACT], data[prl.DONE]

        if not isinstance(actor_dist, torch.distributions.Distribution):
            # If deterministic policy, use action as mean as fix scale to 1.0
            actor_dist = torch.distributions.Normal(loc=a, scale=1.0)

        actor_dist.probs = torch.clamp(actor_dist.probs, self.eps, 1.0 - self.eps)

        kl_div = []
        for behavior, weight in zip(self.behaviors, self.behavior_weights):

            with torch.no_grad():
                _, _, dist_b = behavior.evaluate_actions(o, rhs, d, a)

            if not isinstance(dist_b, torch.distributions.Distribution):
                # If deterministic policy, use action as mean as fix scale to 1.0
                dist_b = torch.distributions.Normal(loc=dist_b, scale=1.0)

            dist_b.probs = torch.clamp(dist_b.probs, self.eps, 1.0 - self.eps)

            div = (kl_divergence(dist_b, actor_dist) - torch.log(weight))

            # div *= torch.exp(- 2 * dist_b.entropy()).detach()

            kl_div.append(div.mean())

        kl_div = min(kl_div)

        return self.loss_term_weight * kl_div, info


class RepulsionKL(PolicyLossAddOn):

    def __init__(self,
                 behavior_factories,
                 behavior_weights,
                 loss_term_weight=1.0,
                 eps=1e-8):
        """
        Class to enforce dissimilarity of any algorithm policy to specified list of behaviors.

        Parameters
        ----------
        behavior_factories : list
            List of methods creating the agent behaviors.
        behavior_weights : list
            List of floats giving relative weight to each agent behavior. All weights should be
            positive. Otherwise AssertionError will be raised.
        loss_term_weight : float
            Weight of the KL term in the algorithm policy loss.
        eps : float
            Lower bound for prob values, used to clip action probs.
        """

        # Check sizes match
        assert len(behavior_factories) == len(behavior_weights)

        self.eps = eps
        self.behaviors = []
        self.loss_term_weight = loss_term_weight
        self.behavior_factories = behavior_factories

        # Check all behavior weights are positive
        assert (np.array(behavior_weights) >= 0.0).all()

        # Normalize behavior_weights
        self.behavior_weights = behavior_weights
        self.behavior_weights /= np.sum(self.behavior_weights)

    def setup(self, actor, device):
        """
        Setup addon module by casting behavior weights to torch tensors and
        initializing agent behaviors.
        """

        self.actor = actor
        self.device = device

        # Cast behavior weights to torch tensors
        self.behavior_weights = [torch.tensor(w).to(device) for w in self.behavior_weights]

        # Create behavior instances
        for b in self.behavior_factories:
            self.behaviors.append(b(self.device))

    def compute_loss_term(self, data, actor_dist, info):
        """
        Calculate and add KL Repulsion loss term.
            1. Calculate KL between actor policy and all behaviors.
            2. Compute weighted sum of KL similarities.
            3. Multiply the result by the loss_term_weight.
            4. Keep sign of the loss term so KL between behaviors is maximized.

        Parameters
        ----------
        actor_dist : torch.distributions.Distribution
            Actor action distribution for actions in data[prl.OBS]
        data : dict
            data batch containing all required tensors to compute loss term.
        info : dict
            Dictionary to store log information.

        Returns
        -------
        attraction_kl_loss_term : torch.tensor
            KL loss term.
        info : dict
            Updated info dict.
        """

        o, rhs, a, d = data[prl.OBS], data[prl.RHS], data[prl.ACT], data[prl.DONE]

        if not isinstance(actor_dist, torch.distributions.Distribution):
            # If deterministic policy, use action as mean as fix scale to 1.0
            actor_dist = torch.distributions.Normal(loc=a, scale=1.0)

        actor_dist.probs = torch.clamp(actor_dist.probs, self.eps, 1.0 - self.eps)

        kl_div = torch.tensor(0.0, dtype=torch.float32).to(self.device)
        for behavior, weight in zip(self.behaviors, self.behavior_weights):

            with torch.no_grad():
                _, _, dist_b = behavior.evaluate_actions(o, rhs, d, a)

            if not isinstance(dist_b, torch.distributions.Distribution):
                # If deterministic policy, use action as mean as fix scale to 1.0
                dist_b = torch.distributions.Normal(loc=dist_b, scale=1.0)

            dist_b.probs = torch.clamp(dist_b.probs, self.eps, 1.0 - self.eps)

            div = kl_divergence(dist_b, actor_dist)

            # div *= torch.exp(- 2 * dist_b.entropy()).detach()

            kl_div += div.mean()

        return -1 * self.loss_term_weight * kl_div, info



import torch
import torch.nn as nn

from pytorchrl.agent.actors.utils import init
from pytorchrl.agent.actors.noise import get_noise
from pytorchrl.agent.actors.feature_extractors.ensemble_layer import EnsembleFC


class Deterministic(nn.Module):
    """
    Deterministic prediction of the mean value mu of a learned action distribtion.

    Parameters
    ----------
    num_inputs : int
        Size of input feature maps.
    num_outputs : int
        Number of dims in output space.
    noise : str
        Type of noise that is added to the predicted mu.
    """

    def __init__(self, num_inputs, num_outputs, noise):
        super(Deterministic, self).__init__()

        init_ = lambda m: init(m, nn.init.orthogonal_, lambda x: nn.init.constant_(x, 0))
        self.action_output = init_(nn.Linear(num_inputs, num_outputs))
        self.noise = get_noise(noise)(num_outputs)

    def forward(self, x, deterministic=True):
        """
        Predict distribution parameters from x (obs features)
        and returns predicted noisy action mu of the distribution and the clipped action [-1, 1].

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        deterministic : bool
            Whether to noise is added to the predicted mu or not.
            
        Returns
        -------
        action: torch.tensor
            Next action sampled.
        clipped_action: torch.tensor
            Next action sampled, but clipped to be within the env action space.
        logp: None
            Returns logp 'None' to have equal output to other distributions.
        entropy_dist: None
            Returns logp 'None' to have equal output to other distributions
        dist : torch.Distribution
            Action probability distribution.
        """
        mu = torch.tanh(self.action_output(x))
        if not deterministic:
            noise = self.noise.sample().to(mu.device)
            mu = mu + noise
        clipped_action = torch.clamp(mu, min=-1, max=1)
            
        return mu, clipped_action, None, None, clipped_action

    def evaluate_pred(self, x, pred):
        """
        Predict distribution parameters from x (obs features)
        and returns predicted mu value of the distribution.
        Ignores the pred input parameter.

        Parameters
        ----------
        x : torch.tensor
            Feature maps extracted from environment observations.
        pred : torch.tensor
            Prediction to evaluate.

        Returns
        -------
        logp : torch.tensor
            Log probability of `pred` according to the predicted
             distribution.
        entropy_dist : torch.tensor
            Entropy of the predicted distribution.
        dist : torch.Distribution
            Action probability distribution.
        """
        pred = torch.tanh(self.action_output(x))
        pred = torch.clamp(pred, min=-1, max=1)

        return None, None, pred

class DeterministicMB(nn.Module):
    """Deterministic ensemble output layer 
    
        Parameters
        ----------
        num_inputs: int
            Size of input feature maps.
        num_outputs: int
            Output size of the gaussian layer.
        ensemble_size: int
            Ensemble size in the output layer.
    """
    def __init__(self, num_inputs: int, num_outputs: int)-> None:
        super(DeterministicMB, self).__init__()

        self.num_outputs = num_outputs
        self.output = nn.Linear(in_features=num_inputs, out_features=num_outputs)
        
    def forward(self, x: torch.Tensor)-> torch.Tensor:
        """Forward pass"""
        mean = self.output(x)
        return mean



from typing import List, Tuple
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as nnf


class LstmNet(nn.Module):
    """Implements a LSTM model."""

    def __init__(self, input_size, layer_size=512, num_layers=3, dropout=0., layer_normalization=False):
        """
        Initializes a N layer LSTM cell.

        Parameters
        ----------
        input_size : int
            Input feature map size.
        output_size : int
            Recurrent hidden state and output size.
        dropout : float
            If non-zero, introduces a Dropout layer on the outputs of each GRU layer except the last layer.
        num_layers : int
            Number of recurrent layers.
        activation : func
            Non-linear activation function.
        layer_normalization : bool
            If True, adds a layer normalization module at the end of the model.
        """
        super(LstmNet, self).__init__()
        self._input_size = input_size
        self._layer_size = layer_size
        self._num_layers = num_layers
        self._dropout = dropout
        self._layer_normalization = layer_normalization
        self._rnn = nn.LSTM(
            self._input_size, self._layer_size, num_layers=self._num_layers,
            dropout=self._dropout, batch_first=False)

    @property
    def num_outputs(self):
        """Output feature map size (as in np.prod(self.output_shape))."""
        return self._layer_size

    @property
    def recurrent_hidden_state_size(self):
        """Recurrent hidden state size"""
        return self._layer_size

    def _forward_lstm(self, x, hxs, done):
        """
        Fast forward pass memory network.

        Parameters
        ----------
        x : torch.tensor
            Feature map obtained from environment observation.
        hxs : torch.tensor
            Current recurrent hidden state.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.

        Returns
        -------
        x : torch.tensor
            Feature map obtained after GRU.
        hxs : torch.tensor
            Updated recurrent hidden state.
        """
        masks = 1 - done
        if x.size(0) == hxs.size(0):
            self._rnn.flatten_parameters()
            x, hxs = self._rnn(x.unsqueeze(0), torch.chunk((torch.transpose(hxs, 0, 1) * masks).contiguous(), 2))
            hxs = torch.transpose(torch.cat(hxs), 0, 1)
            x = x.squeeze(0)

        else:

            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.view(T, N, -1)

            # Same deal with masks
            masks = masks.view(T, N)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = torch.nonzero(((masks[1:] == 0.0).any(dim=-1)), as_tuple=False).squeeze().cpu()

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]
                self._rnn.flatten_parameters()
                rnn_scores, hxs = self._rnn(
                    x[start_idx:end_idx],
                    torch.chunk((torch.transpose(hxs, 0, 1) * masks[start_idx].view(
                        1, -1, 1)).contiguous(), 2))
                hxs = torch.transpose(torch.cat(hxs), 0, 1)
                outputs.append(rnn_scores)

            # x is a (T, N, -1) tensor
            x = torch.cat(outputs, dim=0)

            # flatten
            x = x.view(T * N, -1)

        return x, hxs

    def forward(self, inputs, rhs=None, done=False):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            A tensor containing episode observations.
        rhs : torch.tensor
            A tensor representing the recurrent hidden states.
        done : torch.tensor
            A tensor indicating where episodes end.

        Returns
        -------
        x : torch.tensor
            Output feature map.
        rhs : torch.tensor
            Updated recurrent hidden state.
        """
        x = inputs.view(inputs.size(0), -1)
        x, rhs = self._forward_lstm(x, rhs, done)

        if self._layer_normalization:
            x = nnf.layer_norm(x, x.size()[1:])

        return x, rhs

    def get_initial_recurrent_state(self, num_proc):
        """Returns a tensor of zeros with the expected shape of the model's rhs."""
        return torch.zeros(num_proc, self._num_layers * 2, self._layer_size)



import os
import ray
import torch
import logging
from ray._private.services import get_node_ip_address
from .utils import find_free_port

logger = logging.getLogger(__name__)

default_remote_config = {
    "num_cpus": 1,
    "num_gpus": 0.2,
    "memory": 5 * 1024 ** 3,
    "object_store_memory": 2 * 1024 ** 3
}


class Worker:
    """
    Class containing common worker functionality.

    Parameters
    ----------
    index_worker : int
        Worker index.

    Attributes
    ----------
    index_worker : int
        Index assigned to this worker.
    actor : nn.Module
        An actor class instance.
    """

    def __init__(self, index_worker):
        self.index_worker = index_worker
        self.actor = None  # Initialize in inherited class
        self.is_remote = False

    @classmethod
    def as_remote(cls,
                  num_cpus=None,
                  num_gpus=None,
                  memory=None,
                  object_store_memory=None,
                  resources=None):
        """
        Creates a Worker instance as a remote ray actor.

        Parameters
        ----------
        num_cpus : int
            The quantity of CPU cores to reserve for this Worker class.
        num_gpus  : float
            The quantity of GPUs to reserve for this Worker class.
        memory : int
            The heap memory quota for this actor (in bytes).
        object_store_memory : int
            The object store memory quota for this actor (in bytes).
        resources: Dict[str, float]
            The default resources required by the actor creation task.

        Returns
        -------
        W : Worker
            A ray remote actor Worker class.
        """
        w = ray.remote(
            num_cpus=num_cpus,
            num_gpus=num_gpus,
            memory=memory,
            object_store_memory=object_store_memory,
            resources=resources)(cls)
        w.is_remote = True
        return w

    def print_worker_info(self):
        """Print information about this worker, including index and resources assigned"""
        s = "Created {} with worker_index {}".format(
            str(type(self).__name__), self.index_worker)
        if self.index_worker != 0:
            s += ", in machine {} using gpus {}".format(
                get_node_ip_address(),
                ray.get_gpu_ids())
        logger.warning(s)

    def get_weights(self):
        """Returns current actor.state_dict() weights"""
        return {k: v.cpu() for k, v in self.actor.state_dict().items()}

    def terminate_worker(self):
        """Terminate this ray actor"""
        ray.actor.exit_actor()

    def get_node_ip(self):
        """Returns the IP address of the current node."""
        return get_node_ip_address()

    def find_free_port(self):
        """Returns a free port on the current node."""
        return find_free_port()

    def setup_torch_data_parallel(self, url, rank, world_size, backend):
        """
        Join a torch process group for distributed SGD.

        Parameters
        ----------
        url :
            URL specifying how to initialize the process group.
        rank :
            Rank of the current process.
        world_size : int
            Number of processes participating in the job.
        backend : str
            The pytorch distributed backend to use. valid values include mpi,
            gloo, and nccl.
        """
        torch.distributed.init_process_group(
            backend=backend,
            init_method=url,
            rank=rank,
            world_size=world_size)

    @staticmethod
    def get_host():
        """Return node name where this Worker is being executed."""
        return os.uname()[1]



import os
from pytorchrl.envs.generative_chemistry.reinvent.rnn_environment import GenChemEnv as RnnReinvent
from pytorchrl.envs.generative_chemistry.reinvent.transformer_environment import GenChemEnv as TrReinvent
from pytorchrl.envs.generative_chemistry.vocabulary import SMILESTokenizer, create_vocabulary


def reinvent_train_env_factory(scoring_function, vocabulary, smiles_max_length=200, accumulate_obs=False):
    """
    Create train REINVENT environment.

    Parameters
    ----------
    scoring_function : func
        Function that given a SMILE, returns a score.
    vocabulary : class
        Class that stores the tokens and allows their conversion to vocabulary indexes.
    smiles_max_length : int
        Maximum length allowed for the generated SMILES. Equivalent to max episode length.
    accumulate_obs : bool
        If True, obs are accumulated at every time step (e.g. C, CC, CCC, ...)

    Returns
    -------
    env : gym.Env
        Train environment.
    """

    if accumulate_obs:
        env = TrReinvent(
            scoring_function=scoring_function,
            vocabulary=vocabulary,
            max_length=smiles_max_length)
    else:
        env = RnnReinvent(
            scoring_function=scoring_function,
            vocabulary=vocabulary,
            max_length=smiles_max_length)

    return env



import numpy as np
import torch
import pytorchrl as prl
from pytorchrl.agent.storages.base import Storage as S
import gym
from torch.nn.functional import one_hot


def dim0_reshape(tensor, size):
    """
    Reshapes tensor so indices are defined like this:

    00, 01, 02, 03, 04, 05, 06, 07, 08, 09, size + 1, ..., self.max_size
    10, 11, 12, 13, 14, 15, 16, 17, 18, 19, size + 1, ..., self.max_size
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29, size + 1, ..., self.max_size

    """
    return np.moveaxis(tensor, [0, 1], [1, 0])[:, 0: size].reshape(-1, *tensor.shape[2:])


class MBReplayBuffer(S):
    """Storage class for Model Based algorithms.

    Implements oll necessary functions to handle data storage and processing in model-based RL algorithms.

    Attributes
    ----------
    size : int
        Storage capacity along time axis.
    device : torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    actor : Actor
        Actor class instance.
    algo : Algorithm
        Algorithm class instance
    """

    # Data fields to store in buffer and contained in the generated batches
    storage_tensors = prl.MBDataKeys

    def __init__(self, size, device, actor, algorithm, envs):

        self.actor = actor
        self.scaler = actor.dynamics_model.standard_scaler
        self.device = device
        self.algo = algorithm
        self.max_size, self.size, self.step = size, 0, 0
        self.data = {k: None for k in self.storage_tensors}  # lazy init

        self.reset()

    @classmethod
    def create_factory(cls, size):
        """
        Returns a function that creates ReplayBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.

        Returns
        -------
        create_buffer_instance : func
            creates a new MBReplayBuffer class instance.
        """

        def create_buffer(device, actor, algorithm, envs):
            """Create and return a ReplayBuffer instance."""
            return cls(size, device, actor, algorithm, envs)

        return create_buffer

    def init_tensors(self, sample):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        for k, v in sample.items():

            if k not in self.storage_tensors:
                continue

            if isinstance(v, dict):
                self.data[k] = {}
                for x, y in sample[k].items():
                    self.data[k][x] = np.zeros((self.max_size, *y.shape), dtype=np.float32)
            else:
                self.data[k] = np.zeros((self.max_size, *v.shape), dtype=np.float32)

    def get_data_slice(self, start_pos, end_pos):
        """
        Makes a copy of all tensors in the buffer between steps `start_pos`
        and `end_pos`.

        Parameters
        ----------
        start_pos : int
            initial slice position.
        end_pos : int
            final slice position.

        Returns
        -------
        data : dict
            data slice copied from the buffer.
        """

        copied_data = {k: None for k in self.storage_tensors}

        for k, v in self.data.items():

            if v is None:
                continue

            if isinstance(self.data[k], dict):
                copied_data[k] = {x: None for x in self.data[k]}
                for x, y in v.items():
                    copied_data[k][x] = np.copy(y[start_pos:end_pos])
            else:
                copied_data[k] = np.copy(v[start_pos:end_pos])

        return copied_data

    def get_all_buffer_data(self, data_to_cpu=False):
        """
        Return all currently stored data. If data_to_cpu, no need to do
        anything since data tensors are already in cpu memory.

        Parameters
        ----------
        data_to_cpu : bool
            Whether or not to move data tensors to cpu memory.

        Returns
        -------
        data : dict
            data currently stored in the buffer.
        """

        # Define data structure
        data = {k: None if not isinstance(self.data[k], dict) else
            {x: None for x in self.data[k]} for k in self.data}


        idx = self.step

        # Fill up data
        for k, v in self.data.items():

            if v is None:
                continue

            if isinstance(self.data[k], dict):
                for x, y in self.data[k].items():
                    data[k][x] = y[:idx]
            else:
                data[k] = v[:idx]

        return data

    def reset(self):
        """
        Set class size and step to zero. If self.actor uses RNNs, add overlap
        slice of last sequence before reset at the beginning of the storage.
        """

        self.size -= self.step
        self.step = 0

    def insert_data_slice(self, new_data):
        """
        Appends new_data to currently stored data.

        Parameters
        ----------
        new_data : dict
            Dictionary of env transition samples to be added to self.data.
        """
        lengths = []
        for k, v in new_data.items():

            if v is None:
                continue

            if isinstance(new_data[k], dict):

                if self.data[k] is None:
                    self.data[k] = {i: None for i in new_data[k].keys()}

                for x, y in new_data[k].items():
                    length = self.insert_single_tensor_slice(self.data[k], x, y)
                    lengths.append(length)
            else:
                length = self.insert_single_tensor_slice(self.data, k, v)
                lengths.append(length)

        assert len(set(lengths)) == 1

        self.step = (self.step + length) % self.max_size
        self.size = min(self.size + length, self.max_size)

    def insert_single_tensor_slice(self, tensor_storage, tensor_key, tensor_values):
        """
        Appends tensor_value to buffer dict using tensor_key as key.

        Parameters
        ----------
        tensor_storage :

        tensor_key : str
            key to use to store the tensor.
        tensor_values : np.ndarray
            tensor values.

        Returns
        -------
        l : int
            length (time axe) of the tensor added to the buffer.
        """

        l = tensor_values.shape[0]

        if tensor_storage[tensor_key] is None:  # If not defined, initialize tensor
            tensor_storage[tensor_key] = np.zeros((self.max_size, *tensor_values.shape[1:]), dtype=np.float32)

        if self.step + l <= self.max_size:  # If enough space, add tensor at the end
            tensor_storage[tensor_key][self.step:self.step + l] = tensor_values

        else:  # Circular buffer
            tensor_storage[tensor_key][
            self.step:self.max_size] = tensor_values[0:self.max_size - self.step]
            tensor_storage[tensor_key][0:l - self.max_size + self.step] = tensor_values[self.max_size - self.step:]

        return l

    def insert_transition(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        # Data tensors lazy initialization
        if self.size == 0 and self.data[prl.OBS] is None:
            self.init_tensors(sample)

        # Insert
        for k, v in sample.items():
            if k in self.data.keys():
                if isinstance(sample[k], dict):
                    for x, y in sample[k].items():
                        self.data[k][x][self.step] = y.cpu()
                else:
                    self.data[k][self.step] = v.cpu()

        # Update
        self.step = (self.step + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def before_gradients(self):
        """
        Steps required before updating actor policy model.
        """
        pass

    def after_gradients(self, batch, info):
        """
        Steps required after updating actor policy model validation_percentage
        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """
        return info

    def generate_batches(self, num_mini_batch, mini_batch_size=256, num_epochs=1):
        """
        Returns a batch iterator to update dynamics model.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch. (not used since MB training)
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs. (not used since MB training)

        Yields
        ------
        batch : dict
            Generated data batches.
        """
       
        for k, v in self.data.items():
            if k == prl.OBS:
                observations = v[:self.size]
            elif k == prl.ACT:
                actions = v[:self.size]
            elif k == prl.OBS2:
                next_observations = v[:self.size]
            elif k == prl.REW:
                rewards = v[:self.size]
            else:
                pass

        if type(self.actor.action_space) == gym.spaces.discrete.Discrete:

            actions = actions.astype(np.int64)
            actions = np.eye(actions.max() + 1)[actions].squeeze(1)
            assert actions.shape == (observations.shape[0], 1, self.actor.action_space.n)

        inputs = np.concatenate((observations, actions), axis=-1)
        delta_state = next_observations - observations
        if self.actor.dynamics_model.reward_function is not None:
            targets = delta_state
        else:
            targets = np.concatenate((delta_state, rewards), axis=-1)
        # watch out inputs have shape (all data, 1, dim) not sure why this extra dim
        inputs = torch.from_numpy(inputs).float().to(self.device).squeeze(1)
        targets = torch.from_numpy(targets).float().to(self.device).squeeze(1)

        self.scaler.fit(inputs=inputs, targets=targets)
        norm_inputs, norm_targets = self.scaler.transform(inputs=inputs, targets=targets)

        train_indices = np.arange(observations.shape[0])
        np.random.shuffle(train_indices)
        for batch_number, j in enumerate(range((norm_inputs.shape[0] // mini_batch_size) + 1)):
            start_index = j * mini_batch_size
            indices_shuffled = train_indices[start_index:start_index + mini_batch_size]
            
            input_batch = norm_inputs[indices_shuffled, :]
            target_batch = norm_targets[indices_shuffled, :]
            batch = {"train_input": input_batch,
                     "train_label": target_batch,
                     "batch_number": batch_number}
            yield batch

    def update_storage_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            if parameter_name == "max_size" and self.recurrent_actor:
                new_parameter_value = (new_parameter_value // self.sequence_length) * self.sequence_length
                new_parameter_value *= 2
            setattr(self, parameter_name, new_parameter_value)




from copy import deepcopy
import torch
from torch.utils.data.sampler import BatchSampler, SubsetRandomSampler, SequentialSampler

import pytorchrl as prl
from pytorchrl.utils import RunningMeanStd
from pytorchrl.agent.storages.base import Storage as S


class VanillaOnPolicyBuffer(S):
    """
    Storage class for On-Policy algorithms.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    envs : VecEnv
        Vector of environments instance.
    """

    # Data fields to store in buffer and contained in the generated batches
    storage_tensors = prl.OnPolicyDataKeys

    def __init__(self, size, device, actor, algorithm, envs):

        self.envs = envs
        if self.envs:
            self.num_envs = envs.num_envs
            self.frame_stack, self.frame_skip = 1, 0
            if "frame_stack" in self.envs.env_kwargs.keys():
                self.frame_stack = self.envs.env_kwargs["frame_stack"]
            if "frame_skip" in self.envs.env_kwargs.keys():
                self.frame_skip = self.envs.env_kwargs["frame_skip"]

        self.actor = actor
        self.device = device
        self.algo = algorithm
        self.recurrent_actor = actor.is_recurrent
        self.max_size, self.size, self.step = int(size), 0, 0
        self.data = {k: None for k in self.storage_tensors}  # lazy init

        self.reset()

    @classmethod
    def create_factory(cls, size):
        """
        Returns a function that creates VanillaOnPolicyBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.

        Returns
        -------
        create_buffer_instance : func
            creates a new VanillaOnPolicyBuffer class instance.
        """

        def create_buffer_instance(device, actor, algorithm, envs):
            """Create and return a VanillaOnPolicyBuffer instance."""
            return cls(size, device, actor, algorithm, envs)

        return create_buffer_instance

    def init_tensors(self, sample):
        """
        Lazy initialization of data tensors from a sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """

        for k in sample:

            if k not in self.storage_tensors:
                continue

            if not self.recurrent_actor and k in (prl.RHS, prl.RHS2):
                size = 0
            else:
                size = self.max_size

            # Handle dict sample value
            # init tensors for all dict entries
            if isinstance(sample[k], dict):
                self.data[k] = {}
                for x, v in sample[k].items():
                    self.data[k][x] = torch.zeros(size + 1, *v.shape).to(self.device)

            else:  # Handle non dict sample value
                self.data[k] = torch.zeros(size + 1, *sample[k].shape).to(self.device)

        self.data[prl.RET] = deepcopy(self.data[prl.REW])
        self.data[prl.ADV] = deepcopy(self.data[prl.VAL])

        # In case of an algorithm with intrinsic rewards
        if prl.IREW in sample.keys() and prl.IVAL in sample.keys():
            self.data[prl.IRET] = deepcopy(self.data[prl.IREW])
            self.data[prl.IADV] = deepcopy(self.data[prl.IVAL])
            self.int_reward_rms = RunningMeanStd(shape=(1,), device=self.device)
            self.data[prl.MASK] = torch.ones_like(self.data[prl.DONE])
            self.storage_tensors += (prl.MASK,)

    def get_num_channels_obs(self, sample):
        """
        Obtain num_channels_obs and set it as class attribute.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        try:
            self.num_channels_obs = int(sample[prl.OBS][0].shape[0] // self.frame_stack)
        except (IndexError, KeyError):
            self.num_channels_obs = 1

    def get_all_buffer_data(self, data_to_cpu=False):
        """
        Return all currently stored data.

        Parameters
        ----------
        data_to_cpu : bool
            Whether or not to move data to cpu memory.
        """

        data = {k: v for k, v in self.data.items() if v is not None}

        if data_to_cpu:  # Move tensors to cpu
            for k in data:
                if isinstance(data[k], dict):
                    for x in data[k]:
                        data[k][x] = data[k][x].cpu()
                else:
                    data[k] = data[k].cpu()

        return data

    def insert_data_slice(self, new_data):
        """
        Replace currently stored data.

        Parameters
        ----------
        new_data : dict
            Dictionary of env transition samples to replace self.data with.
        """

        self.data = {k: v for k, v in new_data.items()}

        for k in new_data:

            # Move tensors to self.device
            if isinstance(self.data[k], dict):
                for x in self.data[k]:
                    self.data[k][x] = self.data[k][x].to(self.device)
            else:
                self.data[k] = self.data[k].to(self.device)

    def reset(self):
        """Set class counters to zero and remove stored data"""
        self.step, self.size = 0, 0

    def insert_transition(self, sample):
        """
        Store new transition sample.

        Parameters
        ----------
        sample : dict
            Data sample (containing all tensors of an environment transition)
        """
        if self.size == 0 and self.data[prl.OBS] is None:  # data tensors lazy initialization
            self.init_tensors(sample)
            self.get_num_channels_obs(sample)

        for k in sample:

            if k not in self.storage_tensors:
                continue

            if not self.recurrent_actor and k in (prl.RHS, prl.RHS2):
                continue

            # We use the same tensor to store obs and obs2
            # We also use single tensors for rhs and rhs2,
            # and done and done2
            if k in (prl.OBS, prl.RHS, prl.DONE):
                pos = self.step + 1
                sample_k = "Next" + k
            else:
                pos = self.step
                sample_k = k

            # Copy sample tensor to buffer target position
            if isinstance(sample[k], dict):
                for x, v in sample[k].items():
                    self.data[k][x][pos].copy_(sample[sample_k][x])
            else:
                self.data[k][pos].copy_(sample[sample_k])

        self.step = (self.step + 1) % self.max_size
        self.size = min(self.size + 1, self.max_size)

    def before_gradients(self):
        """
        Before updating actor policy model, compute returns and advantages.
        """

        # Get most recent state
        last_tensors = {}
        step = self.step if self.step != 0 else -1
        for k in (prl.OBS, prl.RHS, prl.DONE):
            if isinstance(self.data[k], dict):
                last_tensors[k] = {x: self.data[k][x][step] for x in self.data[k]}
            else:
                last_tensors[k] = self.data[k][step]

        # Predict values given most recent state
        with torch.no_grad():
            _ = self.actor.get_action(last_tensors[prl.OBS], last_tensors[prl.RHS], last_tensors[prl.DONE])
            value_dict = self.actor.get_value(last_tensors[prl.OBS], last_tensors[prl.RHS], last_tensors[prl.DONE])
            next_rhs = value_dict.get("rhs")

        # Store next recurrent hidden state
        if isinstance(next_rhs, dict):
            for x in self.data[prl.RHS]:
                self.data[prl.RHS][x][step].copy_(next_rhs[x])
        else:
            self.data[prl.RHS][step] = next_rhs

        # Compute returns and advantages
        if isinstance(self.data[prl.VAL], dict):
            # If multiple critics
            for x in self.data[prl.VAL]:
                self.data[prl.VAL][x][step].copy_(value_dict.get(x))
                self.compute_returns(
                    self.data[prl.REW], self.data[prl.RET][x], self.data[prl.VAL][x], self.data[prl.DONE],
                    self.algo.gamma)
                self.data[prl.ADV][x] = self.compute_advantages(self.data[prl.RET][x], self.data[prl.VAL][x])
        else:
            # If single critic
            self.data[prl.VAL][step].copy_(value_dict.get("value_net1"))
            self.compute_returns(
                self.data[prl.REW], self.data[prl.RET], self.data[prl.VAL], self.data[prl.DONE], self.algo.gamma)
            self.data[prl.ADV] = self.compute_advantages(self.data[prl.RET], self.data[prl.VAL])

        # If intrinsic rewards, normalize them
        if hasattr(self.algo, "gamma_intrinsic") and prl.IREW in self.data.keys():
            self.normalize_int_rewards()
            self.algo.state_rms.update(
                self.data[prl.OBS][:, :, -self.num_channels_obs:, ...].reshape(-1, 1, *self.data[prl.OBS].shape[3:]))

        # If algorithm with intrinsic rewards, also compute ireturns and iadvantages
        if prl.IVAL in self.data.keys() and prl.IREW in self.data.keys():
            if isinstance(self.data[prl.IVAL], dict):
                # If multiple critics
                for x in self.data[prl.IVAL]:
                    self.data[prl.IVAL][x][step].copy_(value_dict.get(x))
                    self.compute_returns(
                        self.data[prl.IREW], self.data[prl.IRET][x], self.data[prl.IVAL][x],
                        torch.zeros_like(self.data[prl.DONE]), self.algo.gamma_intrinsic)
                    self.data[prl.IADV][x] = self.compute_advantages(self.data[prl.IRET][x], self.data[prl.IVAL][x])
            else:
                # If single critic
                self.data[prl.IVAL][step].copy_(value_dict.get("ivalue_net1"))
                self.compute_returns(
                    self.data[prl.IREW], self.data[prl.IRET], self.data[prl.IVAL],
                    torch.zeros_like(self.data[prl.DONE]), self.algo.gamma_intrinsic)
                self.data[prl.IADV] = self.compute_advantages(self.data[prl.IRET], self.data[prl.IVAL])

    def after_gradients(self, batch, info):
        """
        After updating actor policy model, make sure self.step is at 0.

        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """

        step = self.step if self.step != 0 else -1
        for k in (prl.OBS, prl.RHS, prl.DONE):
            if isinstance(self.data[k], dict):
                for x in self.data[k]:
                    self.data[k][x][0].copy_(self.data[k][x][step])
            else:
                self.data[k][0].copy_(self.data[k][step])

        if self.step != 0:
            self.step = 0

        return info

    def compute_returns(self, rewards, returns, values, dones, gamma):
        """Compute return values."""
        length = self.step - 1 if self.step != 0 else self.max_size
        returns[length].copy_(values[length])
        for step in reversed(range(length)):
            returns[step] = (returns[step + 1] * gamma * (1.0 - dones[step + 1]) + rewards[step])

    def compute_advantages(self, returns, values):
        """Compute transition advantage values."""
        adv = returns[:-1] - values[:-1]
        adv = (adv - adv.mean()) / (adv.std() + 1e-5)
        return adv

    def normalize_int_rewards(self):
        """
        In order to keep the rewards on a consistent scale, the intrinsic rewards are normalized by dividing
        them by a running estimate of their standard deviation.
        """
        gamma = self.algo.gamma_intrinsic
        length = self.step - 1 if self.step != 0 else self.max_size
        rewems = torch.zeros_like(self.data[prl.RET])
        for step in reversed(range(length)):
            rewems[step] = rewems[step + 1] * gamma + self.data[prl.IREW][step]
        self.int_reward_rms.update(rewems[0:-1].reshape(-1, 1))
        self.data[prl.IREW] = self.data[prl.IREW] / (self.int_reward_rms.var.float() ** 0.5)
        # self.data[prl.IREW] = self.data[prl.IREW] / (self.int_reward_rms.var ** 0.5)

    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1, shuffle=True):
        """
        Returns a batch iterator to update actor.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.
        shuffle : bool
            Whether to shuffle collected data or generate sequential

        Yields
        ------
        batch : dict
            Generated data batches.
        """

        num_proc = self.data[prl.DONE].shape[1]
        l = self.step - 1 if self.step != 0 else self.max_size

        if self.recurrent_actor:  # Batches for a feed recurrent actor

            num_envs_per_batch = num_proc // num_mini_batch
            assert num_proc >= num_mini_batch, "number processes greater than  mini batches"

            for _ in range(num_epochs):
                perm = torch.randperm(num_proc) if shuffle else list(range(num_proc))
                for start_ind in range(0, num_proc, num_envs_per_batch):

                    # Define batch structure
                    batch = {k: [] if not isinstance(self.data[k], dict) else
                    {x: [] for x in self.data[k]} for k in self.data.keys()}

                    # Fill up batch with data
                    for offset in range(num_envs_per_batch):
                        ind = perm[start_ind + offset]
                        for k in batch:
                            # For recurrent state only first position needed
                            len = 1 if k == prl.RHS else l
                            if isinstance(self.data[k], dict):
                                for x in self.data[k]:
                                    batch[k][x].append(self.data[k][x][:len, ind])
                            else:
                                batch[k].append(self.data[k][:len, ind])

                    # Stack and reshape data to target shape
                    for k in batch:
                        if isinstance(self.data[k], dict):
                            for x in self.data[k]:
                                batch[k][x] = torch.stack(batch[k][x], dim=1).view(
                                    -1, *self.data[k][x].size()[2:])
                        else:
                            batch[k] = torch.stack(batch[k], dim=1).view(
                                -1, *self.data[k].size()[2:])

                    yield batch

        else:
            mini_batch_size = num_proc * l // num_mini_batch
            sampler = SubsetRandomSampler if shuffle else SequentialSampler
            for _ in range(num_epochs):
                for samples in BatchSampler(sampler(range(num_proc * l)), mini_batch_size, drop_last=shuffle):

                    batch = {k: None for k in self.data.keys()}

                    for k in batch:

                        if k in (prl.RHS, prl.RHS2):
                            size, idxs = 1, torch.tensor([0])
                        else:
                            size, idxs = l, samples

                        if isinstance(self.data[k], dict):
                            tensor = {x: self.data[k][x][0:size].reshape(
                                -1, *self.data[k][x].shape[2:])[idxs] for x in self.data[k]}
                        else:
                            tensor = self.data[k][0:size].reshape(-1, *self.data[k].shape[2:])[idxs]

                        batch[k] = tensor

                    yield batch

    def update_storage_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)



"""From https://github.com/zuoxingdong/dm2gym/blob/master/dm2gym/envs/dm_suite_env.py"""

import gym
from gym import spaces

from dm_control import suite
from dm_env import specs


def convert_dm_control_to_gym_space(dm_control_space):
    r"""Convert dm_control space to gym space. """
    if isinstance(dm_control_space, specs.BoundedArray):
        space = spaces.Box(
            low=dm_control_space.minimum,
            high=dm_control_space.maximum,
            dtype=dm_control_space.dtype)
        assert space.shape == dm_control_space.shape
        return space
    elif isinstance(dm_control_space, specs.Array) and not isinstance(
            dm_control_space, specs.BoundedArray):
        space = spaces.Box(
            low=-float('inf'),
            high=float('inf'),
            shape=dm_control_space.shape,
            dtype=dm_control_space.dtype)
        return space
    elif isinstance(dm_control_space, dict):
        space = spaces.Dict({key: convert_dm_control_to_gym_space(
            value) for key, value in dm_control_space.items()})
        return space


class DMSuiteEnv(gym.Env):
    def __init__(self, domain_name, task_name, task_kwargs=None, environment_kwargs=None, visualize_reward=False):
        self.env = suite.load(
            domain_name,
            task_name,
            task_kwargs=task_kwargs,
            environment_kwargs=environment_kwargs,
            visualize_reward=visualize_reward)
        self.metadata = {
            'render.modes': ['human', 'rgb_array'],
            'video.frames_per_second': round(1.0 / self.env.control_timestep())
        }
        self.observation_space = convert_dm_control_to_gym_space(self.env.observation_spec())
        self.action_space = convert_dm_control_to_gym_space(self.env.action_spec())
        self.viewer = None

    def seed(self, seed):
        return self.env.task.random.seed(seed)

    def step(self, action):
        timestep = self.env.step(action)
        observation = timestep.observation
        reward = timestep.reward
        done = timestep.last()
        info = {}
        return observation, reward, done, info

    def reset(self):
        timestep = self.env.reset()
        return timestep.observation

    def render(self, mode='human', **kwargs):
        if 'camera_id' not in kwargs:
            kwargs['camera_id'] = 0  # Tracking camera
        use_opencv_renderer = kwargs.pop('use_opencv_renderer', False)

        img = self.env.physics.render(**kwargs)
        if mode == 'rgb_array':
            return img
        elif mode == 'human':
            if self.viewer is None:
                if not use_opencv_renderer:
                    from gym.envs.classic_control import rendering
                    self.viewer = rendering.SimpleImageViewer(maxwidth=1024)
                else:
                    from . import OpenCVImageViewer
                    self.viewer = OpenCVImageViewer()
            self.viewer.imshow(img)
            return self.viewer.isopen
        else:
            raise NotImplementedError

    def close(self):
        if self.viewer is not None:
            self.viewer.close()
            self.viewer = None
        return self.env.close()



import torch


class Normal:
    """ Gaussian Noise """
    def __init__(self, action_size, mean=0, std=0.1):
        self.mean = torch.FloatTensor([mean])
        self.std = torch.FloatTensor([std])
        # TODO: test if sampling per action dimension 
        # improves exploration of one sampled epsilon to 
        # each dimension 
        
    def sample(self):
        return torch.normal(mean=self.mean, std=self.std)
    
    def reset(self,):
        pass
